@@ -34,20 +34,20 @@
 #include <string>
 #include <vector>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <pthread.h>
-#include <stdint.h>
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
 #include <lzlib.h>
 
-#if CHAR_BIT != 8
-#  error "Environments where CHAR_BIT != 8 are not supported."
-#endif
-
 #include "arg_parser.h"
 #include "plzip.h"
 
+#if CHAR_BIT != 8
+#error "Environments where CHAR_BIT != 8 are not supported."
+#endif
+
 #ifndef LLONG_MAX
 #define LLONG_MAX  0x7FFFFFFFFFFFFFFFLL
 #endif
@@ -61,10 +61,10 @@
 
 namespace {
 
+const char * const Program_name = "Plzip";
+const char * const program_name = "plzip";
+const char * const program_year = "2010";
 const char * invocation_name = 0;
-const char * const Program_name    = "Plzip";
-const char * const program_name    = "plzip";
-const char * const program_year    = "2010";
 
 #ifdef O_BINARY
 const int o_binary = O_BINARY;
@@ -77,16 +77,17 @@ struct { const char * from; const char * to; } const known_extensions[] = {
   { ".tlz", ".tar" },
   { 0,      0      } };
 
-struct lzma_options
+struct Lzma_options
   {
   int dictionary_size;		// 4KiB..512MiB
   int match_len_limit;		// 5..273
   };
 
-enum Mode { m_compress = 0, m_decompress, m_test };
+enum Mode { m_compress, m_decompress, m_test };
 
 std::string output_filename;
 int outfd = -1;
+mode_t outfd_mode = S_IRUSR | S_IWUSR;
 bool delete_output_on_interrupt = false;
 pthread_t main_thread;
 pid_t main_thread_pid;
@@ -99,18 +100,16 @@ void show_help() throw()
   std::printf( "\nOptions:\n" );
   std::printf( "  -h, --help                 display this help and exit\n" );
   std::printf( "  -V, --version              output version information and exit\n" );
-//  std::printf( "  -b, --member-size=<n>      set member size limit in bytes\n" );
   std::printf( "  -B, --data-size=<n>        set input data block size in bytes\n" );
   std::printf( "  -c, --stdout               send output to standard output\n" );
   std::printf( "  -d, --decompress           decompress\n" );
   std::printf( "  -f, --force                overwrite existing output files\n" );
   std::printf( "  -k, --keep                 keep (don't delete) input files\n" );
-  std::printf( "  -m, --match-length=<n>     set match length limit in bytes [80]\n" );
+  std::printf( "  -m, --match-length=<n>     set match length limit in bytes [36]\n" );
   std::printf( "  -n, --threads=<n>          set the number of (de)compression threads\n" );
   std::printf( "  -o, --output=<file>        if reading stdin, place the output into <file>\n" );
   std::printf( "  -q, --quiet                suppress all messages\n" );
   std::printf( "  -s, --dictionary-size=<n>  set dictionary size limit in bytes [8MiB]\n" );
-//  std::printf( "  -S, --volume-size=<n>      set volume size limit in bytes\n" );
   std::printf( "  -t, --test                 test compressed file integrity\n" );
   std::printf( "  -v, --verbose              be verbose (a 2nd -v gives more)\n" );
   std::printf( "  -1 .. -9                   set compression level [default 6]\n" );
@@ -141,7 +140,7 @@ void show_version() throw()
   }
 
 
-long long getnum( const char * ptr, const int bs = 0,
+long long getnum( const char * const ptr, const int bs = 0,
                   const long long llimit = LLONG_MIN + 1,
                   const long long ulimit = LLONG_MAX ) throw()
   {
@@ -150,7 +149,7 @@ long long getnum( const char * ptr, const int bs = 0,
   long long result = strtoll( ptr, &tail, 0 );
   if( tail == ptr )
     {
-    show_error( "bad or missing numerical argument", 0, true );
+    show_error( "Bad or missing numerical argument.", 0, true );
     std::exit( 1 );
     }
 
@@ -180,7 +179,7 @@ long long getnum( const char * ptr, const int bs = 0,
       }
     if( bad_multiplier )
       {
-      show_error( "bad multiplier in numerical argument", 0, true );
+      show_error( "Bad multiplier in numerical argument.", 0, true );
       std::exit( 1 );
       }
     for( int i = 0; i < exponent; ++i )
@@ -192,14 +191,14 @@ long long getnum( const char * ptr, const int bs = 0,
   if( !errno && ( result < llimit || result > ulimit ) ) errno = ERANGE;
   if( errno )
     {
-    show_error( "numerical argument out of limits" );
+    show_error( "Numerical argument out of limits." );
     std::exit( 1 );
     }
   return result;
   }
 
 
-int get_dict_size( const char * arg ) throw()
+int get_dict_size( const char * const arg ) throw()
   {
   char *tail;
   int bits = std::strtol( arg, &tail, 0 );
@@ -223,7 +222,7 @@ int extension_index( const std::string & name ) throw()
   }
 
 
-int open_instream( const std::string & name, struct stat * in_statsp,
+int open_instream( const std::string & name, struct stat * const in_statsp,
                    const Mode program_mode, const int eindex,
                    const bool force, const bool to_stdout ) throw()
   {
@@ -231,7 +230,7 @@ int open_instream( const std::string & name, struct stat * in_statsp,
   if( program_mode == m_compress && !force && eindex >= 0 )
     {
     if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: input file `%s' already has `%s' suffix.\n",
+      std::fprintf( stderr, "%s: Input file `%s' already has `%s' suffix.\n",
                     program_name, name.c_str(),
                     known_extensions[eindex].from );
     }
@@ -248,14 +247,16 @@ int open_instream( const std::string & name, struct stat * in_statsp,
       {
       const int i = fstat( infd, in_statsp );
       const mode_t & mode = in_statsp->st_mode;
-      if( i < 0 || !( S_ISREG( mode ) || ( to_stdout &&
-                      ( S_ISFIFO( mode ) || S_ISSOCK( mode ) ||
-                        S_ISBLK( mode ) || S_ISCHR( mode ) ) ) ) )
+      const bool can_read = ( i == 0 &&
+                              ( S_ISBLK( mode ) || S_ISCHR( mode ) ||
+                                S_ISFIFO( mode ) || S_ISSOCK( mode ) ) );
+      if( i != 0 || ( !S_ISREG( mode ) && ( !to_stdout || !can_read ) ) )
         {
         if( verbosity >= 0 )
-          std::fprintf( stderr, "%s: input file `%s' is not a regular file%s.\n",
+          std::fprintf( stderr, "%s: Input file `%s' is not a regular file%s.\n",
                         program_name, name.c_str(),
-                        to_stdout ? "" : " and `--stdout' was not specified" );
+                        ( can_read && !to_stdout ) ?
+                        " and `--stdout' was not specified" : "" );
         close( infd );
         infd = -1;
         }
@@ -286,32 +287,25 @@ void set_d_outname( const std::string & name, const int i ) throw()
     }
   output_filename = name; output_filename += ".out";
   if( verbosity >= 0 )
-    std::fprintf( stderr, "%s: can't guess original name for `%s' -- using `%s'.\n",
+    std::fprintf( stderr, "%s: Can't guess original name for `%s' -- using `%s'.\n",
                   program_name, name.c_str(), output_filename.c_str() );
   }
 
 
 bool open_outstream( const bool force ) throw()
   {
-  if( force )
-    outfd = open( output_filename.c_str(),
-                  O_CREAT | O_TRUNC | O_WRONLY | o_binary,
-                  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
-  else outfd = open( output_filename.c_str(),
-                     O_CREAT | O_EXCL | O_WRONLY | o_binary,
-                     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
-  if( outfd < 0 )
+  int flags = O_CREAT | O_WRONLY | o_binary;
+  if( force ) flags |= O_TRUNC; else flags |= O_EXCL;
+
+  outfd = open( output_filename.c_str(), flags, outfd_mode );
+  if( outfd < 0 && verbosity >= 0 )
     {
-    if( errno == EEXIST ) outfd = -2; else outfd = -1;
-    if( verbosity >= 0 )
-      {
-      if( outfd == -2 )
-        std::fprintf( stderr, "%s: Output file %s already exists, skipping.\n",
-                      program_name, output_filename.c_str() );
-      else
-        std::fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
-                      program_name, output_filename.c_str(), std::strerror( errno ) );
-      }
+    if( errno == EEXIST )
+      std::fprintf( stderr, "%s: Output file `%s' already exists, skipping.\n",
+                    program_name, output_filename.c_str() );
+    else
+      std::fprintf( stderr, "%s: Can't create output file `%s': %s.\n",
+                    program_name, output_filename.c_str(), std::strerror( errno ) );
     }
   return ( outfd >= 0 );
   }
@@ -319,7 +313,7 @@ bool open_outstream( const bool force ) throw()
 
 bool check_tty( const int infd, const Mode program_mode ) throw()
   {
-  if( program_mode == m_compress && isatty( outfd ) )
+  if( program_mode == m_compress && outfd >= 0 && isatty( outfd ) )
     {
     show_error( "I won't write compressed data to a terminal.", 0, true );
     return false;
@@ -356,8 +350,9 @@ void close_and_set_permissions( const struct stat * const in_statsp )
   bool error = false;
   if( in_statsp )
     {
-    if( fchmod( outfd, in_statsp->st_mode ) != 0 ) error = true;
-    else (void)fchown( outfd, in_statsp->st_uid, in_statsp->st_gid );
+    if( fchmod( outfd, in_statsp->st_mode ) != 0 ||
+        ( fchown( outfd, in_statsp->st_uid, in_statsp->st_gid ) != 0 &&
+          errno != EPERM ) ) error = true;
     // fchown will in many cases return with EPERM, which can be safely ignored.
     }
   if( close( outfd ) == 0 ) outfd = -1;
@@ -373,7 +368,7 @@ void close_and_set_permissions( const struct stat * const in_statsp )
     }
   if( error )
     {
-    show_error( "I can't change output file attributes." );
+    show_error( "Can't change output file attributes." );
     cleanup_and_fail( 1 );
     }
   }
@@ -389,15 +384,11 @@ extern "C" void signal_handler( int sig ) throw()
   }
 
 
-void set_signals( const bool to_file ) throw()
+void set_signals() throw()
   {
-  if( to_file )
-    {
-    signal( SIGHUP, signal_handler );
-    signal( SIGINT, signal_handler );
-    signal( SIGTERM, signal_handler );
-    }
-  signal( SIGUSR1, signal_handler );
+  std::signal( SIGHUP, signal_handler );
+  std::signal( SIGINT, signal_handler );
+  std::signal( SIGTERM, signal_handler );
   }
 
 } // end namespace
@@ -406,6 +397,12 @@ void set_signals( const bool to_file ) throw()
 int verbosity = 0;
 
 
+// This can be called from any thread, main thread or sub-threads alike, since
+// they all call common helper functions that call fatal() in case of an error.
+//
+void fatal() { signal_handler( SIGUSR1 ); }
+
+
 void Pretty_print::operator()( const char * const msg ) const throw()
   {
   if( verbosity >= 0 )
@@ -423,40 +420,36 @@ void Pretty_print::operator()( const char * const msg ) const throw()
   }
 
 
-void show_error( const char * msg, const int errcode, const bool help ) throw()
+void show_error( const char * const msg, const int errcode, const bool help ) throw()
   {
   if( verbosity >= 0 )
     {
-    if( msg && msg[0] != 0 )
+    if( msg && msg[0] )
       {
       std::fprintf( stderr, "%s: %s", program_name, msg );
-      if( errcode > 0 ) std::fprintf( stderr, ": %s", std::strerror( errcode ) );
+      if( errcode > 0 )
+        std::fprintf( stderr, ": %s", std::strerror( errcode ) );
       std::fprintf( stderr, "\n" );
       }
-    if( help && invocation_name && invocation_name[0] != 0 )
-      std::fprintf( stderr, "Try `%s --help' for more information.\n", invocation_name );
+    if( help && invocation_name && invocation_name[0] )
+      std::fprintf( stderr, "Try `%s --help' for more information.\n",
+                    invocation_name );
     }
   }
 
 
-void internal_error( const char * msg )
+void internal_error( const char * const msg )
   {
-  std::string s( "internal error: " ); s += msg;
-  show_error( s.c_str() );
+  if( verbosity >= 0 )
+    std::fprintf( stderr, "%s: internal error: %s.\n", program_name, msg );
   std::exit( 3 );
   }
 
 
-// This can be called from any thread, main thread or sub-threads alike, since
-// they all call common helper functions that call fatal() in case of an error.
-//
-void fatal() { signal_handler( SIGUSR1 ); }
-
-
 // Returns the number of bytes really read.
 // If (returned value < size) and (errno == 0), means EOF was reached.
 //
-int readblock( const int fd, uint8_t * buf, const int size ) throw()
+int readblock( const int fd, uint8_t * const buf, const int size ) throw()
   {
   int rest = size;
   errno = 0;
@@ -475,7 +468,7 @@ int readblock( const int fd, uint8_t * buf, const int size ) throw()
 // Returns the number of bytes really written.
 // If (returned value < size), it is always an error.
 //
-int writeblock( const int fd, const uint8_t * buf, const int size ) throw()
+int writeblock( const int fd, const uint8_t * const buf, const int size ) throw()
   {
   int rest = size;
   errno = 0;
@@ -490,26 +483,27 @@ int writeblock( const int fd, const uint8_t * buf, const int size ) throw()
   }
 
 
-int main( const int argc, const char * argv[] )
+int main( const int argc, const char * const argv[] )
   {
   // Mapping from gzip/bzip2 style 1..9 compression modes
   // to the corresponding LZMA compression modes.
-  const lzma_options option_mapping[] =
+  const Lzma_options option_mapping[] =
     {
-    { 1 << 20,  10 },		// -1
-    { 1 << 20,  12 },		// -2
-    { 1 << 20,  17 },		// -3
-    { 1 << 21,  26 },		// -4
-    { 1 << 22,  44 },		// -5
-    { 1 << 23,  80 },		// -6
-    { 1 << 24, 108 },		// -7
-    { 1 << 24, 163 },		// -8
+    { 1 << 20,   5 },		// -0
+    { 1 << 20,   5 },		// -1
+    { 3 << 19,   6 },		// -2
+    { 1 << 21,   8 },		// -3
+    { 3 << 20,  12 },		// -4
+    { 1 << 22,  20 },		// -5
+    { 1 << 23,  36 },		// -6
+    { 1 << 24,  68 },		// -7
+    { 3 << 23, 132 },		// -8
     { 1 << 25, 273 } };		// -9
-  lzma_options encoder_options = option_mapping[5];	// default = "-6"
+  Lzma_options encoder_options = option_mapping[6];	// default = "-6"
   int data_size = 0;
   int debug_level = 0;
   int infd = -1;
-  int num_workers = 0;		// Start this many worker threads
+  int num_workers = 0;		// start this many worker threads
   Mode program_mode = m_compress;
   bool force = false;
   bool keep_input_files = false;
@@ -533,7 +527,8 @@ int main( const int argc, const char * argv[] )
 
   const Arg_parser::Option options[] =
     {
-    { '1', "fast",            Arg_parser::no  },
+    { '0', "fast",            Arg_parser::no  },
+    { '1',  0,                Arg_parser::no  },
     { '2',  0,                Arg_parser::no  },
     { '3',  0,                Arg_parser::no  },
     { '4',  0,                Arg_parser::no  },
@@ -546,6 +541,7 @@ int main( const int argc, const char * argv[] )
     { 'B', "data-size",       Arg_parser::yes },
     { 'c', "stdout",          Arg_parser::no  },
     { 'd', "decompress",      Arg_parser::no  },
+    { 'e', "extreme",         Arg_parser::no  },
     { 'D', "debug",           Arg_parser::yes },
     { 'f', "force",           Arg_parser::no  },
     { 'h', "help",            Arg_parser::no  },
@@ -561,7 +557,7 @@ int main( const int argc, const char * argv[] )
     { 'V', "version",         Arg_parser::no  },
     {  0 ,  0,                Arg_parser::no  } };
 
-  Arg_parser parser( argc, argv, options );
+  const Arg_parser parser( argc, argv, options );
   if( parser.error().size() )				// bad option
     { show_error( parser.error().c_str(), 0, true ); return 1; }
 
@@ -570,20 +566,19 @@ int main( const int argc, const char * argv[] )
     {
     const int code = parser.code( argind );
     if( !code ) break;					// no more options
-    const char * arg = parser.argument( argind ).c_str();
+    const char * const arg = parser.argument( argind ).c_str();
     switch( code )
       {
-      case '1': case '2': case '3':
-      case '4': case '5': case '6':
-      case '7': case '8': case '9':
-                encoder_options = option_mapping[code-'1']; break;
+      case '0': case '1': case '2': case '3': case '4':
+      case '5': case '6': case '7': case '8': case '9':
+                encoder_options = option_mapping[code-'0']; break;
       case 'b': break;
       case 'B': data_size = getnum( arg, 0, 2 * LZ_min_dictionary_size(),
                                     2 * LZ_max_dictionary_size() ); break;
       case 'c': to_stdout = true; break;
       case 'd': program_mode = m_decompress; break;
-      case 'D': debug_level = getnum( arg, 0, 0, 3 );
-                break;
+      case 'D': debug_level = getnum( arg, 0, 0, 3 ); break;
+      case 'e': break;				// ignored by now
       case 'f': force = true; break;
       case 'h': show_help(); return 0;
       case 'k': keep_input_files = true; break;
@@ -601,10 +596,12 @@ int main( const int argc, const char * argv[] )
       case 'V': show_version(); return 0;
       default : internal_error( "uncaught option" );
       }
-    }
+    } // end process options
 
   if( data_size <= 0 )
     data_size = 2 * std::max( 65536, encoder_options.dictionary_size );
+  else if( data_size < encoder_options.dictionary_size )
+    encoder_options.dictionary_size = std::max( data_size, LZ_min_dictionary_size() );
 
   if( num_workers <= 0 )
     {
@@ -622,8 +619,10 @@ int main( const int argc, const char * argv[] )
     }
 
   if( filenames.empty() ) filenames.push_back("-");
-  set_signals( !to_stdout && program_mode != m_test &&
-               ( filenames_given || default_output_filename.size() ) );
+  if( !to_stdout && program_mode != m_test &&
+      ( filenames_given || default_output_filename.size() ) )
+    set_signals();
+  std::signal( SIGUSR1, signal_handler );
 
   Pretty_print pp( filenames );
   if( program_mode == m_test )
@@ -648,6 +647,7 @@ int main( const int argc, const char * argv[] )
           if( program_mode == m_compress )
             set_c_outname( default_output_filename );
           else output_filename = default_output_filename;
+          outfd_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
           if( !open_outstream( force ) )
             {
             if( outfd == -1 && retval < 1 ) retval = 1;
@@ -672,6 +672,7 @@ int main( const int argc, const char * argv[] )
           if( program_mode == m_compress )
             set_c_outname( input_filename );
           else set_d_outname( input_filename, eindex );
+          outfd_mode = S_IRUSR | S_IWUSR;
           if( !open_outstream( force ) )
             {
             if( outfd == -1 && retval < 1 ) retval = 1;
@@ -711,9 +712,7 @@ int main( const int argc, const char * argv[] )
     }
   if( outfd >= 0 && close( outfd ) != 0 )
     {
-    if( verbosity >= 0 )
-      std::fprintf( stderr, "%s: Can't close stdout: %s.\n",
-                    program_name, std::strerror( errno ) );
+    show_error( "Can't close stdout", errno );
     if( retval < 1 ) retval = 1;
     }
   return retval;
